@@ -1,4 +1,4 @@
-__version__ = "0.5.2"
+__version__ = "0.5.10"
 
 # List of names that are part of the public API.
 __all__ = [
@@ -13,7 +13,6 @@ __all__ = [
     "debounce",
     "delayed",
     "expirable",
-    "loggable",
     "log_carrier",
     "pinned",
     "produces",
@@ -31,7 +30,6 @@ __all__ = [
     "DelayedTrait",
     "EvaluatableTrait",
     "ExpirableTrait",
-    "LoggableTrait",
     "LogCarrierTrait",
     "PinnedTrait",
     "ProducesTrait",
@@ -102,7 +100,6 @@ from .traits.decorators import (
     delayed,
     expirable,
     log_carrier,
-    loggable,
     pinned,
     produces,
     rate_limit,
@@ -129,7 +126,7 @@ from .traits.lifecycle import (
 from .traits.lifecycle import (
     QueuedTrait as Queued,
 )
-from .traits.loggable import LogCarrierTrait, LoggableTrait
+from .traits.loggable import LogCarrierTrait
 from .traits.pinned import PinnedTrait
 from .traits.produces import ProducesTrait
 from .traits.rate_limit import RateLimitTrait
