@@ -17,7 +17,7 @@ from typing import (
 import msgpack
 
 from wombat.multiprocessing.errors import EvaluationFailureError
-from wombat.multiprocessing.ipc.utilities import AsyncProcessLock, default_encoder
+from wombat.multiprocessing.ipc.utilities import default_encoder
 from wombat.multiprocessing.traits.accounting import (
     AccountingTrait,
     GeneratedTrait,
@@ -41,7 +41,7 @@ from wombat.multiprocessing.traits.lifecycle import (
     Skipped,
     Succeeded,
 )
-from wombat.multiprocessing.traits.loggable import LoggableTrait, LogCarrierTrait
+from wombat.multiprocessing.traits.loggable import LogCarrierTrait
 from wombat.multiprocessing.traits.pinned import PinnedTrait
 from wombat.multiprocessing.traits.produces import ProducesTrait
 from wombat.multiprocessing.traits.rate_limit import RateLimitTrait
@@ -66,6 +66,7 @@ if TYPE_CHECKING:
     from wombat.multiprocessing.orchestrator import Orchestrator, OrchestratorBuilder
     from wombat.multiprocessing.traits.models import BaseTrait
     from wombat.multiprocessing.worker import Worker
+    from wombat.multiprocessing.ipc.shared_memory_hash_map import SharedMemoryHashMap
 
 
 class BaseSystem:
@@ -75,24 +76,58 @@ class BaseSystem:
     required_traits: ClassVar[List[Type[BaseTrait]]] = []
 
 
-async def _increment_raw_count(worker: "Worker", key: str, value: int = 1):
-    """A raw, non-idempotent counter increment for accounting."""
-    accounting_store = worker.props["accounting_store"].instance
-    worker_key = worker.identity.name
-    async_lock = AsyncProcessLock(accounting_store.lock, worker.runtime_state.loop)
-
-    async with async_lock:
-        data = accounting_store._read_data()
-
-        worker_counts = data.get(worker_key, {})
+def _sync_increment_raw_count(
+    accounting_store: "SharedMemoryHashMap", worker_key: str, key: str, value: int
+):
+    """Synchronous helper to increment worker and total counts under a single lock."""
+    with accounting_store.lock:
+        worker_counts = accounting_store.get(worker_key, {})
         worker_counts[key] = worker_counts.get(key, 0) + value
-        data[worker_key] = worker_counts
+        accounting_store[worker_key] = worker_counts
 
-        total_counts = data.get("Total", {})
+        total_counts = accounting_store.get("Total", {})
         total_counts[key] = total_counts.get(key, 0) + value
-        data["Total"] = total_counts
+        accounting_store["Total"] = total_counts
 
-        accounting_store._write_data(data)
+
+def _sync_increment_worker_raw_count(
+    accounting_store: "SharedMemoryHashMap", worker_key: str, key: str, value: int
+):
+    """Synchronous helper to increment only worker counts under a single lock."""
+    with accounting_store.lock:
+        worker_counts = accounting_store.get(worker_key, {})
+        worker_counts[key] = worker_counts.get(key, 0) + value
+        accounting_store[worker_key] = worker_counts
+
+
+async def _increment_worker_raw_count(worker: "Worker", key: str, value: int = 1):
+    """
+    A raw, non-idempotent counter increment for worker-only accounting, run in
+    an executor to prevent deadlocks.
+    """
+    await worker.runtime_state.loop.run_in_executor(
+        None,
+        _sync_increment_worker_raw_count,
+        worker.props["accounting_store"].instance,
+        worker.identity.name,
+        key,
+        value,
+    )
+
+
+async def _increment_raw_count(worker: "Worker", key: str, value: int = 1):
+    """
+    A raw, non-idempotent counter increment for accounting, run in an executor
+    to prevent deadlocks.
+    """
+    await worker.runtime_state.loop.run_in_executor(
+        None,
+        _sync_increment_raw_count,
+        worker.props["accounting_store"].instance,
+        worker.identity.name,
+        key,
+        value,
+    )
 
 
 class AccountingSystem(BaseSystem):
@@ -110,26 +145,43 @@ class AccountingSystem(BaseSystem):
     @staticmethod
     def before_build(builder: "OrchestratorBuilder", actions: dict[str, Callable]):
         """
-        Auto-provisions a `SharedMemoryDict` for accounting if not already present.
+        Auto-provisions a `SharedMemoryHashMap` for accounting if not already present.
 
         This build hook ensures that the necessary shared memory resource for
         storing task counts is created and managed by the orchestrator.
         """
         if "accounting_store" not in builder.props:
-            from wombat.multiprocessing.ipc.shared_memory_dict import SharedMemoryDict
+            from wombat.multiprocessing.ipc.shared_memory_hash_map import (
+                SharedMemoryHashMap,
+            )
             from wombat.multiprocessing.traits.models import Prop
 
             context = builder.context
-            # Give it a decent size to avoid overflow with many workers/counters
-            shm_dict = SharedMemoryDict.create(
-                context=context, max_size=1024 * 1024, purpose="accounting"
+            # Use a reasonably large capacity and slot size for accounting data.
+            shm_map = SharedMemoryHashMap.create(
+                context=context, capacity=4096, slot_size=1024, purpose="accounting"
             )
-            builder.register_resource_for_cleanup(shm_dict)
+            builder.register_resource_for_cleanup(shm_map)
             builder.props["accounting_store"] = Prop(
-                initializer=shm_dict,
+                initializer=shm_map,
                 use_context_manager=False,
             )
 
+    @staticmethod
+    async def _increment_worker_count(
+        worker: "Worker", accounting_trait: "AccountingTrait", key: str, value: int = 1
+    ):
+        """Idempotent counter increment that only affects the worker's counts."""
+        # Retries are cumulative and not idempotent, so they bypass the `counted` check.
+        if key != "retries":
+            if accounting_trait.counted.get(key):
+                return
+
+        await _increment_worker_raw_count(worker, key, value)
+
+        if key != "retries":
+            accounting_trait.counted[key] = True
+
     @staticmethod
     async def _increment_count(
         worker: "Worker", accounting_trait: "AccountingTrait", key: str, value: int = 1
@@ -144,6 +196,14 @@ class AccountingSystem(BaseSystem):
         if key != "retries":
             accounting_trait.counted[key] = True
 
+    @staticmethod
+    async def on_batch_received(batch: list[Task], worker: "Worker"):
+        """
+        Initial task counting is handled by the `on_task_received` hook to
+        prevent race conditions. This hook is a no-op.
+        """
+        pass
+
     @staticmethod
     @priority(500)
     async def on_task_received(task: "Task", worker: "Worker") -> None:
@@ -160,18 +220,29 @@ class AccountingSystem(BaseSystem):
         if not accounting_trait:
             return
 
+        # Handle 'initial' tasks specially to update only the worker count.
+        # The orchestrator has already updated the 'Total' count.
+        is_initial_task = any(isinstance(t, UncountedTrait) for t in task.traits)
+        if is_initial_task:
+            await AccountingSystem._increment_worker_count(
+                worker, accounting_trait, "initial"
+            )
+
         processed_a_received_event = False
         # Iterate over a copy, as we may modify the list.
         for trait in task.traits[:]:
             if trait.counting_rules:
                 for rule in trait.counting_rules:
                     if rule.on_event == "received":
-                        await AccountingSystem._increment_count(
+                        # This handles 'generated' and 'logs'. The orchestrator now
+                        # handles the 'Total' count, and the worker handles its own count.
+                        await AccountingSystem._increment_worker_count(
                             worker, accounting_trait, rule.counter_name
                         )
                         processed_a_received_event = True
 
-        if processed_a_received_event:
+        if is_initial_task or processed_a_received_event:
+            # Remove the marker trait(s) after counting to prevent re-counting on retries.
             task.remove_traits_by_type(UncountedTrait, GeneratedTrait)
 
     @staticmethod
@@ -248,7 +319,9 @@ class BreakerSystem(BaseSystem):
         This build hook ensures that the `breaker_lock` and `breaker_states`
         props are available for all workers if any task uses the `Breaker` trait.
         """
-        from wombat.multiprocessing.ipc.shared_memory_dict import SharedMemoryDict
+        from wombat.multiprocessing.ipc.shared_memory_hash_map import (
+            SharedMemoryHashMap,
+        )
         from wombat.multiprocessing.traits.models import Prop
 
         context = builder.context
@@ -260,10 +333,10 @@ class BreakerSystem(BaseSystem):
             )
 
         if "breaker_states" not in builder.props:
-            shm_dict = SharedMemoryDict.create(context=context, purpose="breaker")
-            builder.register_resource_for_cleanup(shm_dict)
+            shm_map = SharedMemoryHashMap.create(context=context, purpose="breaker")
+            builder.register_resource_for_cleanup(shm_map)
             builder.props["breaker_states"] = Prop(
-                initializer=shm_dict,
+                initializer=shm_map,
                 use_context_manager=False,
             )
 
@@ -320,7 +393,7 @@ class BreakerSystem(BaseSystem):
                     return False, task  # Block execution
             return True, task
         finally:
-            lock.release()
+            await worker.runtime_state.loop.run_in_executor(None, lock.release)
 
     @staticmethod
     @priority(300)
@@ -365,7 +438,7 @@ class BreakerSystem(BaseSystem):
                 shared_state["failures"] = 0
                 states_dict[key] = shared_state
         finally:
-            lock.release()
+            await worker.runtime_state.loop.run_in_executor(None, lock.release)
         return None
 
     @staticmethod
@@ -418,7 +491,7 @@ class BreakerSystem(BaseSystem):
                     )
             states_dict[key] = shared_state
         finally:
-            lock.release()
+            await worker.runtime_state.loop.run_in_executor(None, lock.release)
 
 
 class DebounceSystem(BaseSystem):
@@ -429,21 +502,23 @@ class DebounceSystem(BaseSystem):
     @staticmethod
     def before_build(builder: "OrchestratorBuilder", actions: dict[str, Callable]):
         """
-        Auto-provisions a `SharedMemoryDict` for deduplication if not already present.
+        Auto-provisions a `SharedMemoryHashMap` for deduplication if not present.
 
         This build hook ensures that the necessary shared memory resource for
         storing task execution timestamps is created and managed by the
         orchestrator.
         """
         if "deduplication_cache" not in builder.props:
-            from wombat.multiprocessing.ipc.shared_memory_dict import SharedMemoryDict
+            from wombat.multiprocessing.ipc.shared_memory_hash_map import (
+                SharedMemoryHashMap,
+            )
             from wombat.multiprocessing.traits.models import Prop
 
             context = builder.context
-            shm_dict = SharedMemoryDict.create(context=context, purpose="debounce")
-            builder.register_resource_for_cleanup(shm_dict)
+            shm_map = SharedMemoryHashMap.create(context=context, purpose="debounce")
+            builder.register_resource_for_cleanup(shm_map)
             builder.props["deduplication_cache"] = Prop(
-                initializer=shm_dict,
+                initializer=shm_map,
                 use_context_manager=False,
             )
 
@@ -491,28 +566,34 @@ class DebounceSystem(BaseSystem):
             return True, task
 
         key = DebounceSystem._get_task_key(task, debounce_trait)
-        now = worker.get_time()
-
         dedupe_cache = worker.props["deduplication_cache"].instance
-        async_lock = AsyncProcessLock(dedupe_cache.lock, worker.runtime_state.loop)
 
-        async with async_lock:
-            data = dedupe_cache._read_data()
-            last_seen = data.get(key)
+        def _check_and_set_sync():
+            """Performs the check and set operation atomically under a lock."""
+            with dedupe_cache.lock:
+                now = worker.get_time()
+                last_seen = dedupe_cache.get(key)
+                if (
+                    last_seen is not None
+                    and (now - last_seen) <= debounce_trait.window.total_seconds()
+                ):
+                    return False
+                dedupe_cache[key] = now
+                return True
 
-            if last_seen and (
-                now - last_seen <= debounce_trait.window.total_seconds()
-            ):
-                worker.log(
-                    f"Task {task.id} (key: {key}) is a duplicate within the {debounce_trait.window} window. Skipping.",
-                    logging.INFO,
-                )
-                task.replace_trait(Skipped())
-                return False, task
+        should_run = await worker.runtime_state.loop.run_in_executor(
+            None, _check_and_set_sync
+        )
 
-            data[key] = now
-            dedupe_cache._write_data(data)
-            return True, task
+        if not should_run:
+            worker.log(
+                f"Task {task.id} (key: {key}) is a duplicate within the {debounce_trait.window} window. Skipping.",
+                logging.INFO,
+            )
+            task.replace_trait(Skipped())
+            return False, task
+
+        return True, task
 
 
 class DelayedSystem(BaseSystem):
@@ -718,44 +799,7 @@ class LifecycleSystem(BaseSystem):
 class LoggableSystem(BaseSystem):
     """System to handle logging for tasks."""
 
-    required_traits: ClassVar[List[Type[BaseTrait]]] = [LoggableTrait, LogCarrierTrait]
-
-    # --- From Loggable ---
-    @staticmethod
-    @priority(800)
-    async def before_task_execution(
-        task: "Task", worker: "Worker"
-    ) -> tuple[bool, "Task"]:
-        """Hook executed before the task's action is run."""
-        if any(isinstance(t, LoggableTrait) for t in task.traits):
-            worker.log(f"Executing task {task.id} ({task.action})", logging.DEBUG)
-        return True, task
-
-    @staticmethod
-    @priority(900)
-    async def on_task_success(
-        task: "Task", worker: "Worker", result: Any
-    ) -> dict[str, Any] | None:
-        """Hook executed after the task's action runs successfully."""
-        loggable_trait = next((t for t in task.traits if isinstance(t, LoggableTrait)), None)
-        if loggable_trait:
-            worker.log(
-                f"Task {task.id} ({task.action}) finished successfully.",
-                loggable_trait.log_level,
-            )
-        return None
-
-    @staticmethod
-    @priority(900)
-    async def on_task_failure(
-        task: "Task", worker: "Worker", exception: Exception | str | None
-    ):
-        """Hook executed when the task's action raises an exception."""
-        if any(isinstance(t, LoggableTrait) for t in task.traits):
-            worker.log(
-                f"Task {task.id} ({task.action}) failed with exception: {exception}",
-                logging.ERROR,
-            )
+    required_traits: ClassVar[List[Type[BaseTrait]]] = [LogCarrierTrait]
 
     # --- From LogCarrier ---
     @staticmethod
@@ -870,14 +914,16 @@ class RateLimitSystem(BaseSystem):
     @staticmethod
     def before_build(builder: "OrchestratorBuilder", actions: dict[str, Callable]):
         if "rate_limit_cache" not in builder.props:
-            from wombat.multiprocessing.ipc.shared_memory_dict import SharedMemoryDict
+            from wombat.multiprocessing.ipc.shared_memory_hash_map import (
+                SharedMemoryHashMap,
+            )
             from wombat.multiprocessing.traits.models import Prop
 
             context = builder.context
-            shm_dict = SharedMemoryDict.create(context=context, purpose="rate_limit")
-            builder.register_resource_for_cleanup(shm_dict)
+            shm_map = SharedMemoryHashMap.create(context=context, purpose="rate_limit")
+            builder.register_resource_for_cleanup(shm_map)
             builder.props["rate_limit_cache"] = Prop(
-                initializer=shm_dict,
+                initializer=shm_map,
                 use_context_manager=False,
             )
 
@@ -893,16 +939,15 @@ class RateLimitSystem(BaseSystem):
             return True, task
 
         rate_limit_cache = worker.props["rate_limit_cache"].instance
-        async_lock = AsyncProcessLock(rate_limit_cache.lock, worker.runtime_state.loop)
 
-        while True:
-            now = worker.get_time()
-            wait_time = 0
-
-            await async_lock.acquire()
-            try:
-                all_timestamps = rate_limit_cache._read_data()
-                group_timestamps = all_timestamps.get(rate_limit_trait.group, [])
+        def _check_rate_limit_sync() -> float:
+            """
+            Checks the rate limit and updates timestamps atomically under a lock.
+            Returns the required wait time in seconds.
+            """
+            with rate_limit_cache.lock:
+                now = worker.get_time()
+                group_timestamps = rate_limit_cache.get(rate_limit_trait.group, [])
 
                 # Prune old timestamps
                 period_seconds = rate_limit_trait.period.total_seconds()
@@ -913,24 +958,26 @@ class RateLimitSystem(BaseSystem):
                 if len(valid_timestamps) < rate_limit_trait.limit:
                     # We are within the limit, add our timestamp and proceed.
                     valid_timestamps.append(now)
-                    all_timestamps[rate_limit_trait.group] = valid_timestamps
-                    rate_limit_cache._write_data(all_timestamps)
-                    return True, task  # Exit the loop and proceed with execution.
+                    rate_limit_cache[rate_limit_trait.group] = valid_timestamps
+                    return 0.0  # No wait time needed.
 
                 # We are at the limit, calculate wait time.
-                # Timestamps are sorted by insertion, so the first is the oldest.
                 oldest_ts = valid_timestamps[0]
-                wait_time = (oldest_ts + period_seconds) - now
-            finally:
-                await async_lock.release()
+                return (oldest_ts + period_seconds) - now
 
-            if wait_time > 0:
-                worker.log(
-                    f"Rate limit for group '{rate_limit_trait.group}' reached. Waiting for {wait_time:.2f}s.",
-                    logging.DEBUG,
-                )
-                await asyncio.sleep(wait_time)
-            # After sleeping, loop again to re-evaluate.
+        while True:
+            wait_time = await worker.runtime_state.loop.run_in_executor(
+                None, _check_rate_limit_sync
+            )
+
+            if wait_time <= 0:
+                return True, task
+
+            worker.log(
+                f"Rate limit for group '{rate_limit_trait.group}' reached. Waiting for {wait_time:.2f}s.",
+                logging.DEBUG,
+            )
+            await asyncio.sleep(wait_time)
 
 
 class RequiresPropsSystem(BaseSystem):
@@ -1012,6 +1059,10 @@ class RetryableSystem(BaseSystem):
 
         if retry_trait.tries < retry_trait.max_tries:
             retry_trait.tries += 1
+            worker.log(
+                f"DEBUG: Scheduling retry {retry_trait.tries}/{retry_trait.max_tries} for task {task.id}",
+                logging.DEBUG,
+            )
             # Use the new trait helpers to robustly manage the traits list.
             task.remove_traits_by_type(UncountedTrait, GeneratedTrait)
             task.replace_trait(retry_trait)
