@@ -18,7 +18,6 @@ from collections.abc import Callable
 from concurrent.futures import ThreadPoolExecutor
 from contextlib import AsyncExitStack
 from enum import Enum
-from queue import Full
 from traceback import format_exc
 from typing import TYPE_CHECKING, Any, Dict, Optional, Union
 
@@ -37,18 +36,12 @@ from wombat.multiprocessing.ipc.utilities import (
     queue_get_async,
 )
 from wombat.multiprocessing.logging import log_task
-from wombat.multiprocessing.systems import AccountingSystem
-from wombat.multiprocessing.traits.accounting import AccountingTrait
 from wombat.multiprocessing.traits.lifecycle import (
     Cancelled,
-    Expired,
-    Failed,
     RetryingTrait,
     Succeeded,
-    Skipped,
 )
 from wombat.multiprocessing.traits.models import (
-    EOQ,
     Prop,
     PropConfig,
     Task,
@@ -91,7 +84,9 @@ class WorkerIPCConfig(BaseModel):
 
 
 class WorkerLoggingConfig(BaseModel):
+    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)
     enabled: bool = True
+    shutdown_event: Optional[Any] = None
 
 
 class WorkerConfig(BaseModel):
@@ -213,19 +208,28 @@ class Worker:
         task.add_done_callback(self.runtime_state.background_tasks.discard)
 
     async def _send_task_result(self, task: "Task") -> None:
-        """Creates a `TaskResult` and puts it on the result queue."""
+        """
+        Creates a `TaskResult` and puts it on the result queue.
+
+        This runs the blocking `put` operation in a thread pool executor to avoid
+        stalling the worker's event loop.
+        """
         result_queue = self.ipc.queues.get("result")
-        if result_queue:
-            task_result = TaskResult(
-                task_id=task.id,
-                action=task.action,
-                result=task.result,
-                metadata=task.metadata,
-                traits=[t.model_dump(mode="python") for t in task.traits],
+        if not result_queue:
+            return
+
+        task_result = TaskResult(
+            task_id=task.id,
+            action=task.action,
+            result=task.result,
+            metadata=task.metadata,
+            traits=[t.model_dump(mode="python") for t in task.traits],
+        )
+
+        if self.runtime_state.loop and self.runtime_state.loop.is_running():
+            await self.runtime_state.loop.run_in_executor(
+                None, result_queue.put_blocking, task_result
             )
-            result_queue.put_blocking(task_result)
-        else:
-            pass
 
     async def add_task_for_retry(self, retry_at: float, task: Task) -> None:
         """
@@ -252,26 +256,22 @@ class Worker:
         queue. The orchestrator will then dispatch this task to a dedicated
         log worker, ensuring that logging I/O does not block task workers.
         """
-        if not self.logging.enabled:
+        if (
+            not self.logging.enabled
+            or (self.logging.shutdown_event and self.logging.shutdown_event.is_set())
+        ):
             return
         requeue_queue = self.ipc.queues.get("requeue")
         if requeue_queue:
-            try:
-                # Log tasks are sent to the central requeue queue to be discovered
-                # and accounted for by the orchestrator.
-                requeue_queue.put_nowait(
-                    log_task(
-                        message=f"worker={self.identity.name} | {message}",
-                        level=level,
-                    )
-                )
-            except Full:
-                # Fallback to stderr to ensure the log message is not lost.
-                print(
-                    f"CRITICAL: Requeue queue for logging is full. Message from worker {self.identity.name}: {message}",
-                    file=sys.stderr,
-                    flush=True,
+            # Log tasks are sent to the central requeue queue to be discovered
+            # and accounted for by the orchestrator. This is a blocking call
+            # to prevent log loss if the queue is full.
+            requeue_queue.put_blocking(
+                log_task(
+                    message=f"worker={self.identity.name} | {message}",
+                    level=level,
                 )
+            )
 
     async def _retry_scheduler_loop(self):
         """A background task that manages and requeues retrying tasks from a min-heap."""
@@ -424,6 +424,7 @@ class Worker:
         props: dict[str, Prop],
         is_async: bool,
     ) -> None:
+        self.log(f"Executing task {task.id} ({task.action})", logging.DEBUG)
         """
         Executes a single task, running it through the full trait lifecycle.
 
@@ -476,17 +477,23 @@ class Worker:
             task.result = "Task execution skipped by pre-flight hook."
             # If a pre-flight hook fails a task (e.g., circuit breaker), we
             # must manually trigger the failure hooks to ensure accounting.
-            if any(isinstance(t, Failed) for t in task.traits):
-                if "on_task_failure" in self.system_registry:
-                    for system_func in self.system_registry["on_task_failure"]:
-                        await system_func(
-                            task, self, "Execution skipped by pre-flight hook"
-                        )
+            if "on_task_failure" in self.system_registry:
+                for system_func in self.system_registry["on_task_failure"]:
+                    await system_func(
+                        task, self, "Execution skipped by pre-flight hook"
+                    )
         except asyncio.CancelledError:
+            # A cancellation (e.g., from Timeout or shutdown) is a terminal state.
+            exception_str = "Task was cancelled."
+            self.log(f"Task {task.id}: {exception_str}", logging.WARNING)
+            last_exception = exception_str
+
+            # Trigger the cancellation hooks for systems like Accounting and Lifecycle.
             if "on_task_cancelled" in self.system_registry:
                 for system_func in self.system_registry["on_task_cancelled"]:
                     await system_func(task, self)
-            raise
+            # Don't re-raise. The cancellation is handled, and the task will be
+            # finalized as 'cancelled' in the finally block.
         except Exception as e:
             last_exception = format_exc()
 
@@ -519,6 +526,7 @@ class Worker:
                     level=logging.ERROR,
                 )
         finally:
+            self.log(f"Finished task {task.id} ({task.action})", logging.DEBUG)
             # A result must be sent for all terminal states. A failed attempt that
             # will be retried is NOT a terminal state, so no result should be sent.
             # This is crucial for the orchestrator's `finish_tasks` to correctly
@@ -528,18 +536,6 @@ class Worker:
             if not is_retrying:
                 # This is a terminal state. Update accounting for states not
                 # covered by the existing success/failure hooks.
-                if any(isinstance(t, (Skipped, Expired)) for t in task.traits):
-                    accounting_trait = next(
-                        (t for t in task.traits if isinstance(t, AccountingTrait)), None
-                    )
-                    if accounting_trait:
-                        counter = "expired" if any(isinstance(t, Expired) for t in task.traits) else "skipped"
-                        # Skipped or Expired are handled here. Cancelled is handled
-                        # by the `on_task_cancelled` hook.
-                        await AccountingSystem._increment_count(
-                            self, accounting_trait, counter
-                        )
-
                 if any(isinstance(t, Succeeded) for t in task.traits):
                     result_for_queue = result_data
                     if isinstance(result_data, Task):
@@ -988,18 +984,23 @@ class Worker:
         finally:
             self.log("Main run loop finally block.", logging.DEBUG)
 
-            # Cancel all running and background tasks to ensure a clean shutdown.
-            all_tasks = list(self.runtime_state.running_tasks) + list(
-                self.runtime_state.background_tasks
-            )
-            if all_tasks:
+            # Wait for all running tasks to complete, but cancel background tasks.
+            running_tasks = list(self.runtime_state.running_tasks)
+            if running_tasks:
                 self.log(
-                    f"Shutdown initiated, cancelling {len(all_tasks)} background/in-flight tasks...",
+                    f"Shutdown initiated, waiting for {len(running_tasks)} in-flight tasks...",
                     logging.INFO,
                 )
-                for task in all_tasks:
+                await asyncio.gather(*running_tasks, return_exceptions=True)
+
+            background_tasks = list(self.runtime_state.background_tasks)
+            if background_tasks:
+                self.log(
+                    f"Cancelling {len(background_tasks)} background tasks.", logging.INFO
+                )
+                for task in background_tasks:
                     task.cancel()
-                await asyncio.gather(*all_tasks, return_exceptions=True)
+                await asyncio.gather(*background_tasks, return_exceptions=True)
 
             if not init_task.done():
                 init_task.cancel()
@@ -1010,9 +1011,11 @@ class Worker:
             for prop in self.props.values():
                 if prop.use_context_manager and prop.exit_stack:
                     await prop.exit_stack.aclose()
-            if self.ipc.queues and "result" in self.ipc.queues:
-                self.log("Putting EOQ on result queue.", logging.DEBUG)
-                self.ipc.queues["result"].put(EOQ())
+                elif hasattr(prop.instance, "close"):
+                    # For non-context-manager props that have a close method (like SharedMemoryHashMap)
+                    prop.instance.close()
+            # EOQ is no longer used for shutdown signaling. The orchestrator's
+            # result collector now exits based on the shutdown event.
             self.log("Worker run finished.", logging.DEBUG)
 
     async def _batch_processing_loop(self, init_task, actions, done_callback):
@@ -1034,6 +1037,11 @@ class Worker:
                 None, self._consume_batch_sync
             )
 
+            if tasks_data:
+                if "on_batch_received" in self.system_registry:
+                    for system_func in self.system_registry["on_batch_received"]:
+                        await system_func(tasks_data, self)
+
             if tasks_data:
                 tasks_to_run.extend(tasks_data)
                 # After processing, release the semaphore so the orchestrator can send more work.
