@@ -42,21 +42,29 @@ def crash_worker_task(_worker: Worker):
 
 
 @pytest_asyncio.fixture
-async def orchestrator() -> Orchestrator:
+async def orchestrator(request) -> Orchestrator:
     """Fixture to provide a started orchestrator and ensure it's shut down."""
+    progress_bar_enabled = request.param
     builder = (
         OrchestratorBuilder()
         .with_workers(num_workers=2)
         .with_actions(
-            [simple_sync_task, simple_async_task, worker_name_task, crash_worker_task]
+            [
+                simple_sync_task,
+                simple_async_task,
+                worker_name_task,
+                crash_worker_task,
+            ]
         )
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     orch = builder.build()
     async with orch:
         yield orch
 
 
+@pytest.mark.parametrize("orchestrator", [True, False], indirect=True)
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
 async def test_orchestrator_e2e_lifecycle(orchestrator: Orchestrator):
@@ -78,9 +86,10 @@ async def test_orchestrator_e2e_lifecycle(orchestrator: Orchestrator):
 
 
 @pytest.mark.parametrize("start_method", ["spawn", "fork", "forkserver"])
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_orchestrator_e2e_with_start_methods(start_method):
+async def test_orchestrator_e2e_with_start_methods(start_method, progress_bar_enabled: bool):
     """Tests the e2e lifecycle with different multiprocessing start methods."""
     if start_method == "fork" and sys.platform == "win32":
         pytest.skip("Fork is not available on Windows")
@@ -91,6 +100,7 @@ async def test_orchestrator_e2e_with_start_methods(start_method):
         .with_workers(num_workers=2)
         .with_actions([simple_sync_task, simple_async_task])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         tasks = [simple_sync_task(1, 2), simple_async_task(3, 4)]
@@ -115,7 +125,8 @@ async def test_orchestrator_e2e_with_start_methods(start_method):
 
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_orchestrator_task_distribution_with_pinning():
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
+async def test_orchestrator_task_distribution_with_pinning(progress_bar_enabled: bool):
     """Tests that the Pinned trait correctly routes tasks to a specific worker."""
     builder = (
         OrchestratorBuilder()
@@ -123,6 +134,7 @@ async def test_orchestrator_task_distribution_with_pinning():
         .with_actions([worker_name_task])
         .with_systems([PinnedSystem])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         pinned_task_def_0 = pinned(worker_name="worker-0")(worker_name_task)
@@ -139,6 +151,7 @@ async def test_orchestrator_task_distribution_with_pinning():
         check.equal(results[pinned_task_1.id].result, "worker-1")
 
 
+@pytest.mark.parametrize("orchestrator", [True, False], indirect=True)
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
 async def test_orchestrator_get_results_clears_buffer(orchestrator: Orchestrator):
@@ -156,13 +169,15 @@ async def test_orchestrator_get_results_clears_buffer(orchestrator: Orchestrator
 
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_orchestrator_handles_worker_crash():
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
+async def test_orchestrator_handles_worker_crash(progress_bar_enabled: bool):
     """Tests that the orchestrator raises WorkerCrashError when a worker dies."""
     builder = (
         OrchestratorBuilder()
         .with_workers(num_workers=1)
         .with_actions([crash_worker_task])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         await orchestrator.add_task(crash_worker_task())
@@ -173,13 +188,15 @@ async def test_orchestrator_handles_worker_crash():
 
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_orchestrator_shutdown_with_context_manager():
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
+async def test_orchestrator_shutdown_with_context_manager(progress_bar_enabled: bool):
     """Tests that the async context manager correctly shuts down the orchestrator."""
     builder = (
         OrchestratorBuilder()
         .with_workers(num_workers=1)
         .with_actions([simple_sync_task])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         await orchestrator.add_task(simple_sync_task(1, 1))
@@ -189,6 +206,7 @@ async def test_orchestrator_shutdown_with_context_manager():
     check.is_true(orchestrator.stopped)
 
 
+@pytest.mark.parametrize("orchestrator", [True, False], indirect=True)
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
 async def test_orchestrator_add_task_and_add_tasks(orchestrator: Orchestrator):
@@ -209,13 +227,15 @@ async def test_orchestrator_add_task_and_add_tasks(orchestrator: Orchestrator):
 
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_orchestrator_with_zero_workers():
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
+async def test_orchestrator_with_zero_workers(progress_bar_enabled: bool):
     """Tests that the orchestrator can run and shut down with zero workers."""
     builder = (
         OrchestratorBuilder()
         .with_workers(num_workers=0)
         .with_actions([simple_sync_task])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         await orchestrator.add_task(simple_sync_task(1, 1))
@@ -230,7 +250,8 @@ async def test_orchestrator_with_zero_workers():
 
 @pytest.mark.asyncio
 @pytest.mark.timeout(20)
-async def test_add_tasks_with_unserializable_arg():
+@pytest.mark.parametrize("progress_bar_enabled", [True, False])
+async def test_add_tasks_with_unserializable_arg(progress_bar_enabled: bool):
     """
     Tests that add_tasks handles unserializable arguments gracefully by returning
     them in the `enqueue_failures` list.
@@ -240,6 +261,7 @@ async def test_add_tasks_with_unserializable_arg():
         .with_workers(num_workers=1)
         .with_actions([simple_sync_task])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     async with builder.build() as orchestrator:
         # Lambda functions are not serializable by the custom encoder.
@@ -254,3 +276,22 @@ async def test_add_tasks_with_unserializable_arg():
         await orchestrator.finish_tasks()
         results = list(orchestrator.get_results())
         check.equal(len(results), 0)
+
+
+@pytest.mark.parametrize("orchestrator", [True, False], indirect=True)
+@pytest.mark.asyncio
+@pytest.mark.timeout(20)
+async def test_orchestrator_handles_large_number_of_tasks(orchestrator: Orchestrator):
+    """Tests the orchestrator's ability to handle a large volume of tasks."""
+    num_tasks = 1000
+    tasks = [simple_sync_task(i, i) for i in range(num_tasks)]
+    await orchestrator.add_tasks(tasks)
+    await orchestrator.finish_tasks()
+
+    results = list(orchestrator.get_results())
+    check.equal(len(results), num_tasks)
+
+    # Verify a few results to be sure
+    results.sort(key=lambda r: r.result)
+    expected_results = sorted([i + i for i in range(num_tasks)])
+    check.equal([r.result for r in results], expected_results)
