@@ -22,19 +22,22 @@ def delayed_action(_worker: Worker) -> str:
 
 
 @pytest_asyncio.fixture
-async def orchestrator() -> Orchestrator:
+async def orchestrator(request) -> Orchestrator:
+    progress_bar_enabled = request.param
     builder = (
         OrchestratorBuilder()
         .with_workers(num_workers=1)
         .with_actions([delayed_action])
         .with_systems([DelayedSystem])
         .without_logging()
+        .with_progress_bar(progress_bar_enabled)
     )
     orch = builder.build()
     async with orch:
         yield orch
 
 
+@pytest.mark.parametrize("orchestrator", [True, False], indirect=True)
 @pytest.mark.asyncio
 @pytest.mark.timeout(5)
 async def test_delayed_task(orchestrator: Orchestrator):
